/*
 * Copyright  2007 Keith Packard <keithp@keithp.com>
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/extensions/Xvlib.h>
#define WIDTH 256
#define HEIGHT 256
char		yuyv      [HEIGHT][WIDTH][2];

/*
 * Paint a rectangle using a solid YUV color. Note how alternate pixels
 * receive U and V data
 */
static void
fill_rect(int x0, int y0, int x1, int y1, char Y, char U, char V)
{
	int		x         , y;
	for (y = y0; y < y1; y++)
		for (x = x0; x < x1; x++) {
			yuyv[y][x][0] = Y;
			if ((x & 1) == 0)
				yuyv[y][x][1] = U;
			else
				yuyv[y][x][1] = V;
		}
}

static void
paint_image(char Y)
{
	int		p         , q;
	for (p = 0; p <= WIDTH - 16; p += 16)
		for (q = 0; q <= HEIGHT - 16; q += 16)
			fill_rect(p, q, p + 16, q + 16, Y, p, q);
}

int
main(int argc, char **argv)
{
	Display        *dpy;
	int		num_adaptors;
	XvAdaptorInfo  *adaptor_info;
	int		screen;
	Window		root   , win;
	int		a;
	XvPortID	port;
	XvImageFormatValues *formats, *format;
	int		nformats;
	int		f;
	XSetWindowAttributes wattr;
	XEvent		event;
	GC		gc;
	XvImage        *image;
	XTextProperty	wm_name, icon_name;
	XSizeHints	sizeHints;
	XWMHints	wmHints;
	int		blink = 0;
	int		flash = 0;
	int		p;
	int		y = 127;
	int		width     , height;
	char           *adaptor_name = NULL;
	int		stopped = 0;

	if (argv[1]) {
		adaptor_name = argv[1];
		printf("Looking for adaptor containing \"%s\"\n", adaptor_name);
	}
	width = WIDTH;
	height = HEIGHT;
	wm_name.value = (unsigned char *)argv[0];
	wm_name.encoding = XA_STRING;
	wm_name.format = 8;
	wm_name.nitems = strlen(wm_name.value) + 1;
	icon_name = wm_name;
	dpy = XOpenDisplay(NULL);
	screen = DefaultScreen(dpy);
	root = RootWindow(dpy, screen);
	/*
	 * Ask the server for all of the available adaptors
	 */
	XvQueryAdaptors(dpy, root, &num_adaptors, &adaptor_info);
	/*
	 * Search the adaptors for an available port
	 */
	port = None;
	for (a = 0; a < num_adaptors && port == None; a++) {
		if (adaptor_name && !strcasestr(adaptor_info[a].name, adaptor_name))
			continue;
		printf("Adapator \"%s\" has %d ports\n",
		       adaptor_info[a].name,
		       adaptor_info[a].num_ports);
		for (p = 0; p < adaptor_info[a].num_ports; p++) {
			if (XvGrabPort(dpy, adaptor_info[a].base_id + p, CurrentTime) == Success) {
				port = adaptor_info[a].base_id + p;
				break;
			}
		}
	}
	/*
	 * Uh-oh, no ports available
	 */
	if (port == None) {
		fprintf(stderr, "No port available on %s\n",
			adaptor_info[a].name);
		exit(1);
	}
	/*
	 * Find a suitable YUYV format for this port
	 */
	formats = XvListImageFormats(dpy, port, &nformats);
	for (f = 0; f < nformats; f++) {
		/*
		 * Check a bunch of port format data; most of this is
		 * probably redundant, and there are probably a few checks
		 * missing that are necessary
		 */
		if (formats[f].type != XvYUV)
			continue;
fprintf(stderr,"type: %s\nFormat: %s (%s)\nHY: %d HU: %d HV: %d\nVY: %d VU: %d VV: %d\n\n"
			, formats[f].type==XvYUV?"YUV":(formats[f].type==XvRGB?"RGB":"Unknown")
			, formats[f].format==XvPacked?"Packed":(formats[f].type==XvPlanar?"Planar":"Unknown")
			, formats[f].component_order
			, formats[f].horz_y_period
			, formats[f].horz_u_period
			, formats[f].horz_v_period
			, formats[f].vert_y_period
			, formats[f].vert_u_period
			, formats[f].vert_v_period);
		if (formats[f].format != XvPacked)
			continue;
		if (formats[f].bits_per_pixel != 16)
			continue;
		if (formats[f].num_planes != 1)
			continue;
		if (formats[f].scanline_order != XvTopToBottom)
			continue;
		if (strcmp(formats[f].component_order, "YUYV") != 0)
			continue;
		format = &formats[f];
//		break;
	}
	/*
	 * Display some of the other format data
	 */
	printf("y_sample_bits %d u_sample_bits %d v_sample_bits %d\n",
	format->y_sample_bits, format->u_sample_bits, format->v_sample_bits);
	printf("component_order: %s\n", format->component_order);
	/*
	 * Create a window
	 */
	wattr.event_mask = ExposureMask | KeyPressMask | StructureNotifyMask;
	win = XCreateWindow(dpy, root, 100, 100, width, height, 0,
			    DefaultDepth(dpy, screen),
			    InputOutput, DefaultVisual(dpy, screen),
			    CWEventMask, &wattr);
	sizeHints.flags = 0;
	wmHints.flags = InputHint;
	wmHints.input = True;
	XSetWMProperties(dpy, win,
			 &wm_name, &icon_name,
			 argv, argc,
			 &sizeHints, &wmHints, 0);
	XMapWindow(dpy, win);
	gc = XCreateGC(dpy, win, 0, NULL);
	/*
	 * Paint some stuff in our data block and create the Xv image using
	 * it
	 */
	image = XvCreateImage(dpy, port, format->id, &yuyv[0][0][0], WIDTH, HEIGHT);
	paint_image(y);
	/*
	 * Paint the whole image on every expose
	 */
	while (XNextEvent(dpy, &event) == Success) {
		char		buffer    [16];
		switch (event.type) {
		case Expose:
			XvPutImage(dpy, port, win, gc, image, 0, 0, WIDTH, HEIGHT, 0, 0, width, height);
			if (blink) {
				y = 255 - y;
				paint_image(y);
				XClearArea(dpy, win, 0, 0, 0, 0, True);
			}
			if (flash) {
				XvStopVideo(dpy, port, win);
				XClearArea(dpy, win, 0, 0, 0, 0, True);
			}
			break;
		case ConfigureNotify:
			width = ((XConfigureEvent *) & event)->width;
			height = ((XConfigureEvent *) & event)->height;
			XClearArea(dpy, win, 0, 0, 0, 0, True);
			break;
		case KeyPress:
			XLookupString((XKeyEvent *) & event, buffer, sizeof(buffer), NULL, NULL);
			switch (buffer[0]) {
			case 'q':
				exit(0);
				break;
			case 'b':
				blink = !blink;
				if (blink)
					y = 0;
				else
					y = 127;
				paint_image(y);
				XClearArea(dpy, win, 0, 0, 0, 0, True);
				break;
			case 'f':
				flash = !flash;
				XClearArea(dpy, win, 0, 0, 0, 0, True);
				break;
			case ' ':
				if (!stopped) {
					XvStopVideo(dpy, port, win);
					stopped = 1;
				} else {
					stopped = 0;
					XClearArea(dpy, win, 0, 0, 0, 0, True);
				}
				break;
			default:
				XClearArea(dpy, win, 0, 0, 0, 0, True);
				break;
			}
		}
	}
}
